/*
 * (c) 2017, SWD Embedded Systems Limited, http://www.kpda.ru
 */


#include "proto.h"


extern int  verbose;
uint32_t    gpio_pbase[] = { OMAP3530_GPIO1_BASE,
                             OMAP3530_GPIO2_BASE,
                             OMAP3530_GPIO3_BASE,
                             OMAP3530_GPIO4_BASE,
                             OMAP3530_GPIO5_BASE,
                             OMAP3530_GPIO6_BASE };


#define GPIO_PIN_NUM        (pin_num - (gpio_num * 32))


gpio_dev_t * am35_hw_init( gpio_dev_t *dev )
{
    am35_gpio_t *am35;
    int         i, j;

    if ( -1 == ThreadCtl( _NTO_TCTL_IO, 0 ) )
    {
        gpio_slogf( "ThreadCtl" );
        return (NULL);
    }

    am35 = calloc( 1, sizeof( am35_gpio_t ) );
    if ( !am35 )
        return (NULL);

    for ( i = 0; i <= 5; i++ )
    {
        am35->gpiobase[i] = mmap_device_io( OMAP3530_GPIO_SIZE, gpio_pbase[i] );
        if ( am35->gpiobase[i] == (uintptr_t)MAP_FAILED )
        {
            gpio_slogf( "mmap_device_io GPIO%d failed.", i );
            for( j = 0; j < i; j++ )
                munmap_device_io( am35->gpiobase[j], OMAP3530_GPIO_SIZE );
            free( am35 );
            return (NULL);
        }

        // Disable interrupts
        out32( am35->gpiobase[i] + GPIO_IRQENABLE_1, 0 );
        out32( am35->gpiobase[i] + GPIO_IRQENABLE_2, 0 );

        // Disable interrupt High/Low level detect
        out32( am35->gpiobase[i] + GPIO_LEVELDETECT0, 0 );
        out32( am35->gpiobase[i] + GPIO_LEVELDETECT1, 0 );
        out32( am35->gpiobase[i] + GPIO_RISINGDETECT, 0 );
        out32( am35->gpiobase[i] + GPIO_FALLINGDETECT, 0 );

        // Reset interrupt status
        out32( am35->gpiobase[i] + GPIO_IRQSTATUS_1, 0xFFFFFFFF );
        out32( am35->gpiobase[i] + GPIO_IRQSTATUS_2, 0xFFFFFFFF );
    }

    dev->hdl = am35;

    return dev;
}


void am35_hw_fini( gpio_dev_t *dev )
{
    am35_gpio_t *am35 = dev->hdl;
    int         i;

    for ( i = 0; i <= 5; i++ )
        munmap_device_io( am35->gpiobase[i], OMAP3530_GPIO_SIZE );

    dev->hdl = NULL;
    free( am35 );
}


int hw_cmd_read( void *hdl, void *buf )
{
    am35_gpio_t     *am35       = hdl;
    int             bytes       = 1;
    uint8_t         pin_num;
    uint8_t         gpio_num;
    gpio_devctl_t   *devmsg_ptr = (gpio_devctl_t *)buf;

    pin_num = devmsg_ptr->cmd_read.pin_num;

    if ( pin_num > 191 )
    {
        gpio_slogf( "GPIO error: Valid GPIO pin number is from range 0 ~ 191." );
        return (0);
    }

    gpio_num = pin_num / 32;

    // check if it's input or output
    if ( in32( am35->gpiobase[gpio_num] + GPIO_OE ) & (1 << GPIO_PIN_NUM) ) // GPIO set as input
    {
        if ( verbose )
            gpio_slogf( "GPIO pin_num %d set as input ", pin_num );
        if ( in32( am35->gpiobase[gpio_num] + GPIO_DATAIN ) & (1 << GPIO_PIN_NUM) )
            devmsg_ptr->cmd_read.data = 1;
        else
            devmsg_ptr->cmd_read.data = 0;
    } else {
        // GPIO set as output
        if ( verbose )
            gpio_slogf( "GPIO pin_num %d set as output ", pin_num );
        if ( in32( am35->gpiobase[gpio_num] + GPIO_DATAOUT ) & (1 << GPIO_PIN_NUM) )
            devmsg_ptr->cmd_read.data = 1;
        else
            devmsg_ptr->cmd_read.data = 0;
    }

    if ( verbose )
    {
        gpio_slogf( "GPIO%d DATAIN REG 0x%X ",     (gpio_num + 1), in32( am35->gpiobase[gpio_num] + GPIO_DATAIN ) );
        gpio_slogf( "GPIO%d PIN %d [%d] data %d ", (gpio_num + 1), GPIO_PIN_NUM, pin_num, devmsg_ptr->cmd_read.data );
    }

    return bytes;
}


int hw_cmd_write( void *hdl, void *buf )
{
    am35_gpio_t     *am35       = hdl;
    int             bytes       = 1;
    uint8_t         pin_num;
    uint8_t         gpio_num;
    uint8_t         value;
    gpio_devctl_t   *devmsg_ptr = (gpio_devctl_t *)buf;

    pin_num = devmsg_ptr->cmd_write.pin_num;
    value   = devmsg_ptr->cmd_write.data;

    if ( pin_num > 191 )
    {
        gpio_slogf( "GPIO error: Valid GPIO pin number is from range 0 ~ 191." );
        return (0);
    }

    /* NOTE: we need to set the pins to GPIO mode for below operation.
     * This is done in startup or the individual drivers by setting the
     * conf_<module>_<pin> register */

    gpio_num = pin_num / 32;

    // check if it's input or output
    if ( in32( am35->gpiobase[gpio_num] + GPIO_OE ) & (1 << GPIO_PIN_NUM) ) // GPIO set as input
    {
        gpio_slogf( "GPIO pin set as input, invalid for write." );
        return (0);
    }

    if ( value )
        out32( am35->gpiobase[gpio_num] + GPIO_SETDATAOUT,   (1 << GPIO_PIN_NUM) );
    else
        out32( am35->gpiobase[gpio_num] + GPIO_CLEARDATAOUT, (1 << GPIO_PIN_NUM) );

    return bytes;
}


int hw_set_input( void *hdl, uint8_t data )
{
    am35_gpio_t     *am35   = hdl;
    uint8_t         pin_num = data;
    uint8_t         gpio_num;
    uint32_t        temp;

    if ( (pin_num > 191) || (pin_num < 0) )
    {
        gpio_slogf( "GPIO error: Valid GPIO pin number is from range 0 ~ 191." );
        return (0);
    }

    gpio_num = pin_num / 32;

    temp = in32( am35->gpiobase[gpio_num] + GPIO_OE );
    out32( am35->gpiobase[gpio_num] + GPIO_OE, temp | (1 << GPIO_PIN_NUM) );

    if ( verbose )
        gpio_slogf( "GPIO%d PIN %d (%d) configured as input.", gpio_num, GPIO_PIN_NUM, pin_num );

    return (0);
}


int hw_set_output( void *hdl, uint8_t data )
{
    uint32_t        temp;
    am35_gpio_t     *am35   = hdl;
    uint8_t         pin_num = data;
    uint8_t         gpio_num;

    if ( (pin_num > 191) || (pin_num < 0) )
    {
        gpio_slogf( "GPIO error: Valid GPIO pin number is from range 0 ~ 191." );
        return (-1);
    }

    gpio_num = pin_num / 32;

    // check if it's input or output
    if ( !(in32( am35->gpiobase[gpio_num] + GPIO_OE ) & (1 << GPIO_PIN_NUM)) ) // GPIO set as output
    {
        if ( verbose )
            gpio_slogf( "GPIO pin already set as output." );
        return (0);
    } else {
        temp = in32( am35->gpiobase[gpio_num] + GPIO_OE );
        out32( am35->gpiobase[gpio_num] + GPIO_OE, temp & (~(1 << GPIO_PIN_NUM)) );
    }

    return (0);
}


int hw_int_enable_wlevel( void *hdl, uint8_t data, uint8_t int_level )
{
    uint32_t        temp;
    am35_gpio_t     *am35   = hdl;
    uint8_t         pin_num = data;
    uint8_t         gpio_num;

    if ( (pin_num > 191) || (pin_num < 0) )
    {
        gpio_slogf( "GPIO error: Valid GPIO pin number is from range 0 ~ 191." );
        return (-1);
    }

    gpio_num = pin_num / 32;

    if ( int_level == GPIO_INT_LEVEL_LOW )
    {
        temp = in32( am35->gpiobase[gpio_num] + GPIO_LEVELDETECT0 );
        out32( am35->gpiobase[gpio_num] + GPIO_LEVELDETECT0, temp | (1 << GPIO_PIN_NUM) );
    } else
        if ( int_level == GPIO_INT_LEVEL_HIGH )
        {
            temp = in32( am35->gpiobase[gpio_num] + GPIO_LEVELDETECT1 );
            out32( am35->gpiobase[gpio_num] + GPIO_LEVELDETECT1, temp | (1 << GPIO_PIN_NUM) );
        }

    // Reset interrupt status
    out32( am35->gpiobase[gpio_num] + GPIO_IRQSTATUS_1, (1 << GPIO_PIN_NUM) );
    out32( am35->gpiobase[gpio_num] + GPIO_IRQSTATUS_2, (1 << GPIO_PIN_NUM) );

    out32( am35->gpiobase[gpio_num] + GPIO_SETIRQENABLE1, (1 << GPIO_PIN_NUM) );

    if ( verbose )
        gpio_slogf( "GPIO%d PIN %d (%d) interrupt is enabled.", gpio_num, GPIO_PIN_NUM, pin_num );

    if ( InterruptUnmask( OMAP3530_GPIO_6_IRQ, 0 ) == -1 )
    {
        gpio_slogf( "GPIO%d PIN %d (%d) interrupt enable failed.", gpio_num, GPIO_PIN_NUM, pin_num );
        if ( verbose )
            perror( "InterruptUnmask" );
        return (-1);
    }

    return (0);
}


int hw_int_enable( void *hdl, uint8_t data )
{
    am35_gpio_t     *am35   = hdl;
    uint8_t         pin_num = data;
    uint8_t         gpio_num;

    if ( (pin_num > 191) || (pin_num < 0) )
    {
        gpio_slogf( "GPIO error: Valid GPIO pin number is from range 0 ~ 191." );
        return (-1);
    }

    gpio_num = pin_num / 32;

    out32( am35->gpiobase[gpio_num] + GPIO_SETIRQENABLE1, (1 << GPIO_PIN_NUM) );
    if ( verbose )
        gpio_slogf( "GPIO%d PIN %d (%d) interrupt is enabled.", gpio_num, GPIO_PIN_NUM, pin_num );

    if ( InterruptUnmask( OMAP3530_GPIO_6_IRQ, 0 ) == -1 )
    {
        gpio_slogf( "GPIO%d PIN %d (%d) interrupt unmask failed.", gpio_num, GPIO_PIN_NUM, pin_num );
        if ( verbose )
            perror( "GPIO RM: InterruptUnmask" );
        return (-1);
    }

    return (0);
}


int hw_int_disable( void *hdl, uint8_t pin_num )
{
    am35_gpio_t     *am35 = hdl;
    uint8_t         gpio_num;
    unsigned        temp;

    if ( (pin_num > 191) || (pin_num < 0) )
    {
        gpio_slogf( "GPIO error: Valid GPIO pin number is from range 0 ~ 191." );
        return (-1);
    }

    gpio_num = pin_num / 32;

    // Disable interrupt
    out32( am35->gpiobase[gpio_num] + GPIO_CLEARIRQENABLE1, (1 << GPIO_PIN_NUM) );

    // Disable interrupt High/Low level detect
    temp = in32( am35->gpiobase[gpio_num] + GPIO_LEVELDETECT0 );
    out32( am35->gpiobase[gpio_num] + GPIO_LEVELDETECT0, temp & (~(1 << GPIO_PIN_NUM)) );

    temp = in32( am35->gpiobase[gpio_num] + GPIO_LEVELDETECT1 );
    out32( am35->gpiobase[gpio_num] + GPIO_LEVELDETECT1, temp & (~(1 << GPIO_PIN_NUM)) );

    // Reset interrupt status
    out32( am35->gpiobase[gpio_num] + GPIO_IRQSTATUS_1, (1 << GPIO_PIN_NUM) );
    out32( am35->gpiobase[gpio_num] + GPIO_IRQSTATUS_2, (1 << GPIO_PIN_NUM) );

    if ( verbose )
        gpio_slogf( "GPIO%d PIN %d (%d) interrupt is disabled.", gpio_num, GPIO_PIN_NUM, pin_num );

    return (0);
}


int hw_irq_status( void *hdl, gpio_interrupt_t *gpio_int )
{
    am35_gpio_t     *am35 = hdl;
    uint32_t        temp,
                    i,
                    j;
    uint8_t         pin   = 0,
                    level = 0;

    for ( i = 0; i < 6; i++ )
    {
        temp = in32( am35->gpiobase[i] + GPIO_IRQSTATUS_1 );

        if ( temp != 0x0 )
        {
            if ( verbose )
                gpio_slogf( "GPIO IRQ status 0x%X ", temp );

            // out32( am35->gpiobase[i] + GPIO_IRQSTATUS_1, temp );

            for ( j = 0; j < 32; j++ )
            {
                if ( (temp & 0x1) == 0x1 )
                {
                    pin  = j;
                    break;
                } else
                    temp = temp >> 1;
            }

            if ( in32( am35->gpiobase[i] + GPIO_LEVELDETECT0 ) & (1 << pin) )
                level = GPIO_INT_LEVEL_LOW;
            else if ( in32( am35->gpiobase[i] + GPIO_LEVELDETECT1 ) & (1 << pin) )
                level = GPIO_INT_LEVEL_HIGH;
            else
                level = -1;

            // out32( am35->gpiobase[i] + GPIO_IRQSTATUS_1, 1 << pin );
            pin = pin + (i * 32);
            hw_int_disable( hdl, pin);

            gpio_int->pin   = pin;
            gpio_int->level = level;

            return pin;
        }
    }

    return (-1);
}
